import logging
import os


class NSLogger():
    def __init__(self):
        """
        This class provides an interface to the python logging module.  This
        class also provides methods to view several NetworkScanner data
        structures.
        """
        # Create logger
        self.logger = logging.getLogger(__name__)

        # Default to INFO
        self.logger.setLevel(logging.INFO)

        # Log to the root of the package so data is not mixed with source
        log_file_name = os.path.join("..", "..", "network_scanner.log")
        file_handler = logging.FileHandler(log_file_name)
        format_string = '%(asctime)s : %(levelname)s : %(name)s : %(message)s'
        formatter = logging.Formatter(format_string)
        file_handler.setFormatter(formatter)

        # Add file handler to logger
        self.logger.addHandler(file_handler)

        # Also send logger information to the console
        console = logging.StreamHandler()
        console.setLevel(logging.INFO)
        self.logger.addHandler(console)

    # Wrap 3 of the 5 logging levels: debug, info, and warning.
    # For simplicity, levels error and critical are not used in this package.
    # Errors are reported through exception handling.
    def debug(self, message):
        self.logger.debug(message)

    def info(self, message):
        self.logger.info(message)

    def warning(self, message):
        self.logger.warning(message)


if __name__ == "__main__":
    print("Testing NSLogger")
    ns_log = NSLogger()
    ns_log.info("Logging INFO level message to module log file.")



from network_scanner_kbk import NetworkScanner

print ("Testing NetworkScanner...")

last_round = None
#for i in range(0, 1):
for i in range(0, 4):
    subnet1 = "192.168.0.0/24"
    subnet2 = "192.168.1.0/24"
    
    # subnet1 = "192.168." + str(i%2) + ".0/24"
    # subnet2 = "192.168." + str(2*i + 1) + ".0/24"
    print("\n\n\n" + subnet1 + subnet2)
    retries = 0
    #ignore_list = [23, 45, -1, 256]
    ignore_list = []
    
    nv = NetworkScanner()
    #nodes = nv.xxx(subnet1, subnet2, retries, ignore_list)
    nodes = nv.find_singularly_accesible_nodes(subnet1, subnet2, retries, ignore_list)


    print("Singularly accesible nodes:")
    print(nodes)
    
    print("Subnet1:")
    for key in sorted(nv.ping_status[0]):
        s = "%s: %s" % (key, nv.ping_status[0][key])
        print(s.ljust(10), end ="")
    
    print("Subnet2:")
    total = 0
    for key in sorted(nv.ping_status[1]):
        s = "%s: %s" % (key, nv.ping_status[1][key])
        if nv.ping_status[1][key] == True:
            total += 1
        print(s.ljust(10), end ="")
            
    
    print("\n\nTOTAL: " + str(total))
    if last_round is not None:
        for key in sorted(nv.ping_status[1]):
            if nv.ping_status[1][key] != last_round[key]:
                if last_round[key] == True:
                    print("DISAPPEARED: " + str(key))
                else:
                    print("APPEARED: " + str(key))
    last_round = nv.ping_status[1]     
        
    #print ("Done Testing NetworkScanner...")


    


import subprocess
from threading import Thread
from collections import OrderedDict

from .ns_logger import NSLogger

class NetworkScanner():
    def __init__(self):
        """
        This class provides a library of functions to view network
        properties on the Joby Challenge Problem Device Network.
        """
        # Contains host prefixes for each subnet (subnet1 and subnet2)
        self.host_prefix = [None, None]

        # Pingability status for each host on subnet1, keyed by octet value
        ping_status_subnet1 = OrderedDict()
        # Pingability status for each host on subnet2, keyed by octet value
        ping_status_subnet2 = OrderedDict()
        # List with 1 entry for each subnet
        self.ping_status = [ping_status_subnet1, ping_status_subnet2]

        self.OCTET_MAX_VALUE = 255
        self.MAX_PING_RETRIES = 3
        self.SUBNET1_IDX = 0
        self.SUBNET2_IDX = 1

        self.ns_log = NSLogger()  # Network Scanner Log

    def validate_subnet(self, subnet_input):
        """
        Validate that the subnet_input is valid and return the host
        prefix portion of the subnet_input.  eg. given "192.168.1.0/24,
        return: "192.168.1." (including the trailing ".")
        """
        error_msg = "Unrecognized format for subnet input: " + subnet_input

        if not subnet_input.endswith("/24"):
            raise ValueError(error_msg)
        subnet = subnet_input.replace("/24", "")

        octets = subnet.split(".")
        if len(octets) != 4:
            raise ValueError(error_msg)

        for o in octets:
            if not (o.isnumeric() and int(o) >= 0 and
                    int(o) <= self.OCTET_MAX_VALUE):
                raise ValueError(error_msg)

        host_prefix = subnet_input[0:subnet_input.rfind(".") + 1]
        return host_prefix

    def validate_ignore_list(self, ignore_list):
        """
        Validate that all the values in ignore_list are valid octet values.
        Return a validated ignore_list with invalid octets removed.
        """
        validated_ignore_list = []

        for i in ignore_list:
            if i >= 0 and i <= self.OCTET_MAX_VALUE:
                validated_ignore_list.append(i)
            else:
                self.ns_log.warning("Invalid octet in ignore_list: " + str(i))

        return validated_ignore_list

    def validate_retries(self, retries):
        """ Validate retries is in valid range for this package. """
        if not (retries >= 0 and retries <= self.MAX_PING_RETRIES):
            raise ValueError("Invalid value for retries: " + str(retries))

    def find_singularly_reachable_devices(self, subnet1, subnet2,
                                          retries=1, ignore_list=[]):
        """
        For every device in the network, find all devices that are reachable
        via exactly one subnet.

        Parameters:
            subnet1 (str):ip address for subnet1 in CIDR /24 notation
            subnet2 (str):ip address for subnet2 in CIDR /24 notation
            retries (int):number of additional times to attempt ping
            ignore_list (list):list of octets/host_ids to ignore each subnet

        Returns:
            ips_by_octet(list):List of singularily accesible ips

        """

        # Validate all input arguments
        self.host_prefix[self.SUBNET1_IDX] = self.validate_subnet(subnet1)
        self.host_prefix[self.SUBNET2_IDX] = self.validate_subnet(subnet2)
        validated_ignore_list = self.validate_ignore_list(ignore_list)
        self.validate_retries(retries)

        # For each octet, ping the same octet on each subnet, alternating
        # alternating between subnets to balance traffic accross subnets
        self.ns_log.info("Launching ping threads")
        thread_ids = []
        for octet in range(self.OCTET_MAX_VALUE):
            if octet not in validated_ignore_list:
                for idx in [self.SUBNET1_IDX, self.SUBNET2_IDX]:
                    t = Thread(target=self.ping, args=(octet, idx, retries,))
                    t.start()
                    thread_ids.append(t)
        self.ns_log.info("All threads started..")

        # Wait for all ping threads to complete
        for t in thread_ids:
            t.join()
        self.ns_log.info("All threads completed")

        return self.create_ip_dictionaries()

    def ping(self, octet, subnet_idx, retries):
        tries = retries + 1
        ip = self.host_prefix[subnet_idx] + str(octet)
        subprocess_cmd = 'ping -n ' + str(tries) + ' -w 3000 ' + ip
        self.ns_log.debug("ping called with subnet_idx, subprocess_cmd: "
                          + str(subnet_idx) + ", " + str(subprocess_cmd))

        sp = subprocess.run(subprocess_cmd, capture_output=True, text=True)
        stdout = sp.stdout

        # Simplest way to determine if ip is reachable is if the ping command
        # output contains "time=" followed by ping time in milliseconds.
        # ping -c has better output but requires admin privedlges to run.
        # The return code from ping does not seem to be a reliable indicator
        # of device presence.
        returncode = ""
        if "time=" in stdout:
            returncode = True

        self.ping_status[subnet_idx][octet] = returncode

    def create_ip_dictionaries(self):
        """
        Use the class data to form a list of dictionaries where each dictionary
        contains connectivity information for a single octect.  Octets with
        similar connetivity states are ommited from the dictionary.  Example:

        [
            {
                "192.168.1.25" : true,
                "192.168.1.25" : false
            },
            {
                "192.168.1.27" : false,
                "192.168.1.257" : true
            }
        ]
        """
        ips_by_octet = []

        for octet in sorted(self.ping_status[self.SUBNET1_IDX]):
            # Get the pingability status of octet for each subnet
            subnet1_octet_pingable = self.ping_status[self.SUBNET1_IDX][octet]
            subnet2_octet_pingable = self.ping_status[self.SUBNET2_IDX][octet]

            # If the given octet has a different pingability status on each
            # subnet, add ip and pingability status to output data structure
            if subnet1_octet_pingable != subnet2_octet_pingable:
                subnet1_ip = self.host_prefix[self.SUBNET1_IDX] + str(octet)
                subnet2_ip = self.host_prefix[self.SUBNET2_IDX] + str(octet)

                dict_entry = {subnet1_ip: subnet1_octet_pingable,
                              subnet2_ip: subnet2_octet_pingable}
                self.ns_log.debug("dict_entry: " + str(dict_entry))
                ips_by_octet.append(dict_entry)

        return ips_by_octet


if __name__ == "__main__":
    print("Testing NetworkScanner...")

    from ns_logger import NSLogger
    ns = NetworkScanner()

    subnet1 = "192.168.0.0/24"
    subnet2 = "192.168.1.0/24"
    retries = 0
    ignore_list = [1, 256]
    print("\n\n" + str(subnet1).ljust(15) + str(subnet2).ljust(15)
          + str(retries).ljust(15) + str(ignore_list).ljust(15))

    devices = ns.find_singularly_reachable_devices(subnet1, subnet2,
                                                   retries, ignore_list)

    print("List of devices reachable on exactly one subnet:")
    for d in devices:
        print(d)
    


"""

This script provides an example of using the network_scanner utility to
scan 2 given subnets and look for devices that can only be accessed through
exactly one of the subnets.

"""

# Import the NetworkScanner class
from network_scanner_kbk import NetworkScanner

# Create an instance of the NetworkScanner class
ns = NetworkScanner()

# Provide 2 subnet addresses in CIDR /24 notation
subnet1 = "192.168.0.0/24"
subnet2 = "192.168.1.0/24"

# Select a list of devices/nodes to avoid pinging
ignore_list = [1, 100, 101, 102, 103]

# Call find_singularly_reachable_devices() to obtain the list of devices (ips)
devices = ns.find_singularly_reachable_devices(subnet1, subnet2,
                                               ignore_list=ignore_list)

print("List of devices accessible on exactly one subnet:")
for d in devices:
    print(d)

# This package comes with unit tests.  They can be run as follows.



from setuptools import find_packages
import os
from distutils.core import setup

# User-friendly description from README.md
current_directory = os.path.dirname(os.path.abspath(__file__))
try:
    with open(os.path.join(current_directory, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()
except Exception:
    long_description = ''

setup(
	# Name of the package 
	name='network_scanner_kbk',
    package_dir = {'': 'src'},
	# Packages to include into the distribution 
	packages=["network_scanner_kbk", "network_scanner_kbk.tests"],
	# Start with a small number and increase it with 
	# every change you make https://semver.org 
	version='0.0.19',
	# Chose a license from here: https: // 
	# help.github.com / articles / licensing - a - 
	# repository. For example: MIT 
	license='MIT License',
	# Short description of your library 
	description='Project to scan and gather connectivity information on small networks',
	# Long description of your library 
	long_description=long_description,
	long_description_content_type='text/markdown',
	# Your name 
	author='Brian Kulig',
	# Your email 
	author_email='kbriankulig@gmail.com',
	# Either the link to your github or to your website 
	url='https://github.com/kbriankulig/network-scanner',
	# Link from which the project can be downloaded 
	download_url='',
	# List of keywords 
	keywords=[],
	# List of packages to install with this one 
	install_requires=[],
	# https://pypi.org/classifiers/ 
	classifiers=[]
)
